<?php

namespace Drupal\communication\Form;

use Drupal\communication\Entity\Communication;
use Drupal\communication\OperationPluginManager;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class CommunicationOperationsForm
 *
 * @package Drupal\communication\Form
 */
class CommunicationOperationsForm extends FormBase {

  /**
   * @var \Drupal\communication\OperationPluginManager
   */
  protected $operationManager;

  /**
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * @var \Drupal\communication\Entity\Communication
   */
  protected $communication;

  /**
   * @var array
   */
  protected $options = [
    'exclude' => [],
    'configuration' => [],
  ];

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('plugin.manager.communication.operation'),
      $container->get('renderer')
    );
  }

  /**
   * CommunicationOperationsForm constructor.
   *
   * @param \Drupal\communication\OperationPluginManager $operation_plugin_manager
   */
  public function __construct(OperationPluginManager $operation_plugin_manager, RendererInterface $renderer) {
    $this->operationManager = $operation_plugin_manager;
    $this->renderer = $renderer;
  }

  /**
   * Returns a unique string identifying the form.
   *
   * The returned ID should be a unique string that can be a valid PHP function
   * name, since it's used in hook implementation names such as
   * hook_form_FORM_ID_alter().
   *
   * @return string
   *   The unique string identifying the form.
   */
  public function getFormId() {
    return 'communication_operations_form';
  }

  /**
   * Form constructor.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param \Drupal\communication\Entity\Communication $communication
   *   The communication entity to perform the operation on.
   * @param array $options
   *   Options governing the behaviour of this form. Has the following keys:
   *   - exclude: A list of operation ids. If an operation id appears in this list
   *              it will be excluded from the operation options.
   *   - configuration: Default configuration to provide to the operation.
   *
   * @return array
   *   The form structure.
   */
  public function buildForm(array $form, FormStateInterface $form_state, Communication $communication = NULL, array $options = []) {
    $this->communication = $communication;
    $this->options = $options + $this->options;

    // @todo: Dynamic id to allow for multiple operation forms on one page.
    $form['#prefix'] = '<div id="communication-operations-form">';
    $form['#suffix'] = '</div>';

    $ajax_defaults = [
      'callback' => [$this, 'formAjax'],
      'wrapper' => 'communication-operations-form',
      'method' => 'replace',
    ];

    if (!$form_state->get('operation')) {
      $form['operations'] = [
        '#type' => 'container',
      ];

      /** @var \Drupal\communication\Plugin\Communication\Operation\OperationInterface $plugin */
      foreach ($this->operationManager->availableOperations($communication, $this->options) as $id => $plugin) {
        $form['operations']['do_'.$id] = [
          '#type' => 'submit',
          '#name' => 'do_'.$id,
          '#value' => $plugin->getPluginDefinition()['label'],
          '#submit' => [
            '::formSubmitOperationSelect',
          ],
          '#validate' => [
            '::formValidateOperationSelect',
          ],
          '#ajax' => $ajax_defaults,
          '#operation' => $id,
        ];
      }
    }
    else {
      $operation_id = $form_state->get('operation');
      /** @var \Drupal\communication\Plugin\Communication\Operation\OperationInterface $operation */
      $operation = $this->operationManager->createInstance($operation_id);
      $form = $operation->form($form, $form_state, $communication);

      $form['actions'] = [
        '#type' => 'actions',
      ];
      $form['actions']['submit'] = [
        '#type' => 'submit',
        '#value' => new TranslatableMarkup('Continue'),
        '#submit' => [
          '::formSubmitOperationDo',
        ],
        '#validate' => [
          '::formValidateOperationDo',
        ],
        '#ajax' => $ajax_defaults,
      ];
      $form['actions']['cancel'] = [
        '#type' => 'submit',
        '#value' => new TranslatableMarkup('Cancel'),
        '#submit' => [
          '::formSubmitOperationCancel',
        ],
        '#validate' => [],
        '#ajax' => $ajax_defaults,
      ];
    }

    return $form;
  }

  /**
   * Form submission handler.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // TODO: Implement submitForm() method.
  }

  /**
   * Form ajax.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function formAjax(array $form, FormStateInterface $form_state) {
    $element = $form_state->getTriggeringElement();
    $array_parents = $element['#array_parents'];
    array_pop($array_parents);
    array_pop($array_parents);
    $form_chunk = NestedArray::getValue($form, $array_parents);

    if (!$form_state->get('ajax_commands')) {
      return $form_chunk;
    }
    else {
      $response = new AjaxResponse();
      $response->addCommand(new ReplaceCommand('#communication-operations-form', $this->renderer->render($form_chunk)));
      foreach ($form_state->get('ajax_commands') as $command) {
        $response->addCommand($command);
      }

      return $response;
    }
  }

  /**
   * Validate the operation selection.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function formValidateOperationSelect(array $form, FormStateInterface $form_state) {
    $element = $form_state->getTriggeringElement();
    $operation_id = $element['#operation'];
    /** @var \Drupal\communication\Plugin\Communication\Operation\OperationInterface $plugin */
    $plugin = $this->operationManager->createInstance($operation_id);

    if (!$plugin->applicable($this->communication, $this->options['configuration'])) {
      $form_state->setError($element, new TranslatableMarkup(
        '@operation is not applicable to the communication.',
        [
          '@operation' => $plugin->getPluginDefinition()['label'],
        ]
      ));
    }

    // If this plugin does not have a form, then we need to validate that the
    // operation can be run.
    if (!$plugin->hasForm($this->communication, $this->options['configuration'])) {
      $reasons = [];
      if (!$plugin->validate($this->communication, $this->options['configuration'], $reasons)) {
        foreach ($reasons as $reason) {
          $form_state->setError($element, $reason);
        }
      }
    }
  }

  /**
   * Submit when an operation is selected. If the selected operation has no form
   * then run the operation.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function formSubmitOperationSelect(array $form, FormStateInterface $form_state) {
    $element = $form_state->getTriggeringElement();
    $operation_id = $element['#operation'];
    /** @var \Drupal\communication\Plugin\Communication\Operation\OperationInterface $plugin */
    $plugin = $this->operationManager->createInstance($operation_id);

    if ($plugin->hasForm($this->communication, $this->options['configuration'])) {
      $form_state->set('operation', $operation_id);
      $form_state->setRebuild(TRUE);
    }
    else {
      $plugin->run($this->communication, $this->options['configuration']);
      $plugin->formPostOperation($form, $form_state, $this->communication);
    }
  }

  /**
   * Validate when an operation form has been submitted.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function formValidateOperationDo(array $form, FormStateInterface $form_state) {
    $communication = clone $this->communication;

    /** @var \Drupal\communication\Plugin\Communication\Operation\OperationInterface $plugin */
    $plugin = $this->operationManager->createInstance($form_state->get('operation'));
    $plugin->formValidate($form, $form_state, $communication);

    $initial_options = $form_state->get('operation_options') ?: $this->options['configuration'];
    $form_state->set('operation_options', $initial_options);
    $plugin->formSubmit($form, $form_state, $communication);
    $reasons = [];
    if (!$plugin->validate($communication, $form_state->get('operation_options'), $reasons)) {
      foreach ($reasons as $reason) {
        $form_state->setError($form_state->getTriggeringElement(), $reason);
      }
    }
    $form_state->set('operation_options', $initial_options);
  }

  /**
   * Do the operation.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function formSubmitOperationDo(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\communication\Plugin\Communication\Operation\OperationInterface $plugin */
    $plugin = $this->operationManager->createInstance($form_state->get('operation'));
    $initial_options = $form_state->get('operation_options') ?: $this->options['configuration'];
    $form_state->set('operation_options', $initial_options);
    $plugin->formSubmit($form, $form_state, $this->communication);

    $plugin->run($this->communication, $form_state->get('operation_options'));

    $plugin->formPostOperation($form, $form_state, $this->communication);

    $form_state->set('operation', NULL);
    $form_state->set('operation_options', $this->options['configuration']);
    $form_state->setRebuild(TRUE);
  }

  /**
   * Cancel the operation selected.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   */
  public function formSubmitOperationCancel(array $form, FormStateInterface $form_state) {
    $form_state->set('operation', NULL);
    $form_state->set('operation_options', []);
    $form_state->setRebuild(TRUE);
  }
}
